<?php
/* Copyright 2007-2009 Xavier 'Xr' Dalem
 * 
 * This file is part of Demyo. Demyo is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License version 3,
 * as published by the Free Software Foundation.
 * 
 * Demyo is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU GPLv3 for more details.
 */
/** Defines import/export capabilities for the library.
 * The calling controller should have access to all models (i.e. have them
 * in its $uses variable).
 */
class ImportExportComponent extends Object {
	public $components = array('File', 'CSVImporter');
	private $controller;
	private $errors;
	
	public function startup(&$controller) {
		$this->controller =& $controller;
		$this->errors = array();		
	}
	
	/** Logs an error message.
	 *
	 * @param string $message The message to log. Will be localized.
	 * @return boolean false (always)
	 */
	private function error($message) {
		$this->errors[] = __($message, true);
		return false;
	}
	
	/** Returns a log of errors that occured during processing.
	 *
	 * @return string
	 */
	public function getErrors() {
		return $this->errors;
	}
	
	/** Exports the database.
	 *
	 * @param string $method The export type.
	 * @param boolean $withResources true if associated resources (most notably
	 * images) should be packed with the data.
	 * @return An array with mimetype and content or zip file info, depending
	 * on wether the resources have been included or not. False in case of
	 * error, see {@link #getErrors} for the log.
	 */
	public function export($method, $withResources = false) {
		// Just in case we would need it, act before $method is modified
		$retZipName = "demyo-export-$method.zip";
		
		$exported = array();
		$options = array();
		
		// Detect method
		switch ($method) {
			case 'xml+xsl':
				$options['XSL'] = true;
				// Fall-through
			case 'xml':
				$exported['mimetype'] = 'application/xml';
				$method = 'to_xml';
				break;
			default:
				return $this->error('Invalid export method');
		}
		
		// Run export
		if (! $ret = $this->$method($options))
			return $this->error('Export method failed');
		list($filename, $content) = $ret;
		
		// Return data, or build a zip with the resources
		if (! $withResources)
			return array_merge($exported, array('content' => $content));
		
		// Reset the mimetype: we'll be sending a zip no matter what
		$exported['mimetype'] = 'application/zip';

		$zipName = tempnam(TMP, 'demyo-export-');
		unlink($zipName); // ZipArchive needs a .zip 
		$zip = $zip = new ZipArchive();
		$zipName .= '.zip';
		if ($zip->open($zipName, ZipArchive::CREATE) !== true)
			return $this->error(__('Failed to create ZIP archive.'));;
		
		$zip->addFromString($filename, $content);
		$this->File->addDirToZip($zip, COLLECTION_IMAGES);
		
		$zip->close();
		//register_shutdown_function('unlink', $zipName); // Delete on exit, done from the view (meh, Cake doesn't look adapted for readfile transfers)
		
		return array_merge($exported, array('zipFileName' => $retZipName, 'zipPath' => $zipName));
	}
	
	public function import($method, $formFileData, $forceImport = false) {
		if (file_exists(IMPORT_STATUS_FILE))
			return $this->error('Import already in progress');
		else
			touch(IMPORT_STATUS_FILE);
		
		switch ($method) {
			case 'xml':
				$method = 'from_xml';
				break;
			case 'csv':
				$method = 'from_csv';
				break;
			default:
				return $this->error('Invalid import method');
		}

		if ($formFileData['error'] !== UPLOAD_ERR_OK)
			return $this->error('Upload failed with <a href="http://php.net/manual/en/features.file-upload.errors.php">error code '.$formFileData['error'].'</a>.');

		$ext = pathinfo($formFileData['name'], PATHINFO_EXTENSION);
		
		$parseable = $formFileData['tmp_name'];
		
		if ($ext === 'zip') { // Unzip
			
			// Make temp folder
			$tmpFolder = TMP.DS.'zip_tmp';
			if (file_exists($tmpFolder) && !$this->File->rm($tmpFolder))
				return $this->error('Failed to remove old archive directory');
			if (! mkdir($tmpFolder, 0700))
				return $this->error('Failed to make archive directory');
				
			// Unzip
			$zip = new ZipArchive();
			if ($zip->open($formFileData['tmp_name']) !== true)
				return $this->error('Failed to open zip archive');
			if (! $zip->extractTo($tmpFolder))
				$this->error('Failed to extract archive');
			$zip->close();
			
			// Set import file
			switch ($method) {
				case 'from_xml':
					$parseable = $tmpFolder.DS.'demyo.xml';
					break;
				default:
					return $this->error('No parseable file for archive');
			}
		}
		
		try {
			$ret =  $this->$method($parseable, $forceImport);
		} catch (SchemaMismatchException $e) {
			unlink(IMPORT_STATUS_FILE);
			throw $e;
		}
		
		if ($ret === true && $ext == 'zip') { // Restore images
			
			// Delete all files but root
			if (! $this->File->rm(COLLECTION_IMAGES, true))
				return $this->error('Failed to delete old images');
			
			// Move contents from temporary extraction to final directory
			if (! $this->File->moveDirContent($tmpFolder.DS.COLLECTION_IMAGES_DIR, COLLECTION_IMAGES))
				return $this->error('Failed to restore images');
			// Give rights in final directory. See bug #91
			if (! $this->File->chmod(COLLECTION_IMAGES, 0666))
				return $this->error('Failed to set rights for images');
		}
		
		unlink(IMPORT_STATUS_FILE);
		if (isset($tmpFolder))
			$this->File->rm($tmpFolder);
		return $ret;
	}
	
	public function getImportStatus() {
		if (is_readable(IMPORT_STATUS_FILE))
			return file_get_contents(IMPORT_STATUS_FILE);
		else
			return 'null';
	}
	
	private function to_xml($options) {
		$withXSL = isset($options['XSL']) && $options['XSL'];
		
		// Prepare header before trying to export (fast-fail)
		$ret = "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n";
		$style = '';
		if ($withXSL) {
			$ret .= <<<XSL
<?xml-stylesheet href="#style" type="text/xsl"?>
<!DOCTYPE doc [
<!ATTLIST xsl:stylesheet id ID #REQUIRED>
]>

XSL;
			$style = file_get_contents(WWW_ROOT.DS.'files'.DS.'demyo.xsl');
			if ($style === false)
				return $this->error('Failed to open stylesheet.');
		}
		
		
		// Build XML
		$output = new SimpleXMLElement('<?xml version="1.0" encoding="UTF-8"?><library demyo-version="'
			.Configure::read('Demyo.versionAsFloat').'" schema-version="'.Configure::read('Demyo.schemaVersion').'">'.$style.'</library>');
		$totalEntries = 0;
		foreach (array('Publisher', 'Collection', 'Binding', 'Author', 'Tag',
			'Image', 'Series', 'Album', 'DerivativeType', 'Source', 'Derivative',
			'Borrower', 'AlbumsBorrower') as $model) {
		
			$totalEntries += $this->controller->$model->addToElement($output);
		}
		$output->addAttribute("entries", $totalEntries);
		
		// Format XML
		$doc = new DOMDocument('1.0', 'utf-8');
		$doc->formatOutput = true;
		$domnode = dom_import_simplexml($output);
		$domnode = $doc->importNode($domnode, true);
		$domnode = $doc->appendChild($domnode);
		
		return array('demyo.xml', $ret.$doc->saveXML($domnode));
	}
	
	private function from_xml($parseable, $forceImport = false) {
		if (($xml = simplexml_load_file($parseable)) === false)
			return $this->error('Failed to load XML');
		$xml = current($xml->xpath('//library'));
		
		$attributes = $xml->attributes();
		$entries = $attributes['entries'];
		
		if (!$forceImport
				&& (empty($attributes['schema-version'])
					|| $attributes['schema-version'] < Configure::read('Demyo.schemaVersion'))) {
			throw new SchemaMismatchException();
		}
		
		$reporter = new ProgressReporter($entries, IMPORT_STATUS_FILE);
		$reporter->setAutoWrite(true);
		$processed = 0;
		
		// When importing, the order does matter: the database potentially
		// checks references 
		foreach (array('Publisher', 'Collection', 'Binding', 'Author', 'Tag',
				'Image', 'Series', 'Album', 'DerivativeType', 'Source',
				'Derivative', 'Borrower', 'AlbumsBorrower')
				as $modelName) {
			
			if (!$this->controller->$modelName->importXML($xml, $reporter))
				return false;
		}
		
		return true;
	}
	
	private function from_csv($parseable, $forceImport = false) {
		$reporter = new ProgressReporter(-1, IMPORT_STATUS_FILE);
		$reporter->setAutoWrite(true);
		
		$warnings = $this->CSVImporter->import($this->controller, $parseable, $reporter);
		if (count($warnings) <= 0)
			return true;
		$this->errors = array_merge($this->errors, $warnings);
		return false;
	}
}

class ProgressReporter {
	/** Number of handled entries.
	 * @var int */
	private $current;
	
	/** Total entries to handle.
	 * @var int */
	private $total;
	
	/** Parts progress
	 * @var array
	 */
	private $parts;
	private $currentPart;
	private $outFile;
	private $autoWrite;
	
	function __construct($total, $outFile) {
		$this->current = 0;
		$this->total = $total;
		$this->parts = array();
		$this->currentPart = -1;
		$this->outFile = $outFile;
		$this->autoWrite = false;
	}
	
	function setAutoWrite($auto) {
		$this->autoWrite = $auto;
		if ($auto)
			$this->writeJSON();
	}
	
	function setCurrent($current) {
		$this->current = $current;
		if ($this->autoWrite)
			$this->writeJSON();
	}
	
	function createPart($name, $total) {
		$cur = ++$this->currentPart;
		$this->parts[$cur] = array('name' => $name, 'current' => 0, 'total' => $total);
		if ($this->autoWrite)
			$this->writeJSON();
	}
	
	/** Sets the current amount for the part.
	 * 
	 * @param $current int
	 * @param $updateOverallCurrent bool True if the overall current count
	 * should be adapted (newOverall = oldOverall + (newPart - oldPart))
	 */
	function setPartCurrent($current, $updateOverallCurrent = true) {
		if ($updateOverallCurrent)
			$this->current = $this->current - $this->parts[$this->currentPart]['current'] + $current;
		
		$this->parts[$this->currentPart]['current'] = $current;
		
		if ($this->autoWrite)
			$this->writeJSON();
	}
	
	function getPartCurrent() {
		return $this->parts[$this->currentPart]['current'];
	}
	
	function toJSON() {
		if (($cur = $this->currentPart) > -1) {
			$json = <<<HEREDOC
{
	current: {$this->current},
	total: {$this->total},
	part: "{$this->parts[$cur]['name']}",
	partCurrent: {$this->parts[$cur]['current']},
	partTotal: {$this->parts[$cur]['total']}
}
HEREDOC;
		} else {
			$json = <<<HEREDOC
{
	current: {$this->current},
	total: {$this->total},
	part: null,
	partCurrent: 0,
	partTotal: -1
}
HEREDOC;
		}
		return $json;
	}
	
	function writeJSON() {
		file_put_contents($this->outFile, $this->toJSON());
	}
}

/** Denotes a mismatch between the version of the schema of an import source
 * and the one of the current application.
 */
class SchemaMismatchException extends Exception {
}
?>
