<?php
/* Copyright 2007-2009 Xavier 'Xr' Dalem
 * 
 * This file is part of Demyo. Demyo is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License version 3,
 * as published by the Free Software Foundation.
 * 
 * Demyo is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU GPLv3 for more details.
 */
class SimpleXMLBehavior extends ModelBehavior {
	public $settings;

	public function setup(&$model, $config = array()) {
		$this->settings[$model->name]['useHABTM'] =
			(isset($config['useHABTM']) && $config['useHABTM']);
		$this->settings[$model->name]['container'] =
			empty($config['container'])?
				self::modelNameToContainer($model->name):
				$config['container'];
		$this->settings[$model->name]['element'] =
			empty($config['element'])?
				self::modelNameToElement($model->name):
				$config['element'];
	}
	
	/** Adds a node defining this element to the given SimpleXMLElement
	 * @return int The number of added elements.
	 */
	public function addToElement(&$model, SimpleXMLElement &$element) {
		/* Find out wether we activate habtm handling for this model
		 * (only one of the two models in such a relation needs proper handling
		 */
		$habtm = $this->settings[$model->name]['useHABTM'];
		
		// Container for all "real" records
		$container = $element->addChild($this->settings[$model->name]['container']);
		$model->recursive = $habtm?1:0;
		$list = $model->findAll();
		$container->addAttribute("entries", $entries = count($list));
		
		$nodeName = $this->settings[$model->name]['element'];
		
		// For each record in the base
		foreach ($list as $item) {
			// Create an element for it
			$element = $item[$model->name];
			$child = $container->addChild($nodeName);
			
			// Add trivial attributes (incl. trivial associations)
			foreach ($element as $field => $value)
				$child->addAttribute($field, $value);
			
			// HABTM has to generate a list of all associated records
			if ($habtm && isset($model->hasAndBelongsToMany)) {
				
				// For each HABTM association this model has with another one
				foreach($model->hasAndBelongsToMany as $alias => $assoc) {
					// Find out how many records are associated
					$numIDs = sizeof($item[$alias]);
					$i = 0;
					$value = '';
					
					// Generate a comma-separated list of them
					foreach ($item[$alias] as $otherModelData) {
						$value .= $otherModelData['id'];
						if ($i++ < $numIDs-1)
							$value .= ',';
					}
					
					// Add the brand new attribute
					$child->addAttribute(
						self::HABTMModelToAttribute($alias),
						$value);
				}
			}
		}
		
		return $entries;
	}

	/** Erases records associated with the model and adds those found in
	 * with the given XPath expression
	 *
	 * @param AppModel $model The model 
	 * @param SimpleXMLElement $xml An XML element where records are stored
	 * (e.g. the root element but not the document root)
	 * @param ProgressReporter $reporter
	 */
	public function importXML(&$model, &$xml, $reporter) {
		$xpath = '//'.$this->settings[$model->name]['container']
			.'/'.$this->settings[$model->name]['element'];
		
		$containerAttributes = $xml->{$this->settings[$model->name]['container']}->attributes();
		$reporter->createPart(strtolower(Inflector::humanize(Inflector::underscore(Inflector::pluralize($model->name)))), $containerAttributes['entries']);
		
		if ($this->settings[$model->name]['useHABTM'])
			return $this->importHABTM(&$model, &$xml, $xpath, $reporter);
		else
			return $this->importTrivial(&$model, &$xml, $xpath, $reporter);
	}
	
	/** Trivial import resolves hasOne and belongsTo relationships
	 * (hasMany is resolved by the matching model's belongsTo).
	 * @param ProgressReporter $reporter
	 */
	private function importTrivial(&$model, &$xml, $xpath, $reporter) {
		$imported = 0;
		
		$model->deleteAll('1 = 1', false);
		foreach ($xml->xpath($xpath) as $item) {
			foreach($item->attributes() as $key => $value)
				$data[$model->name][$key] = (string)$value;
			$model->create();
			if (!$model->save($data, false))
				return false;
			
			$imported++;
			if (!($imported%50))
				$reporter->setPartCurrent($imported);
		}
		
		$reporter->setPartCurrent($imported);
		return true;
	}
	
	/** Advanced import resolves all relationships, including HABTM for models
	 * set to work with it.
	 */
	private function importHABTM(&$model, &$xml, $xpath, $reporter) {
		$imported = 0;
		
		// Delete model records
		$model->deleteAll('1 = 1', false);
		
		// Delete model relationships
		foreach ($model->hasAndBelongsToMany as $habtm)
			$model->query('DELETE FROM '.$model->tablePrefix.$habtm['joinTable']);
		
		// HABTM relationships are encoded in CSV (it's about as XML as SVG but
		// it's easier and less verbose)
		
		$aliases = array_keys($model->hasAndBelongsToMany);
		// Attribute names should be derived from aliases, not from classnames (it makes more sense for self-joins)
		$attribNames = array_map(array('SimpleXMLBehavior', 'HABTMModelToAttribute'), $aliases);
		$numAttribs = sizeof($attribNames);
		
		foreach ($xml->xpath($xpath) as $item) {
			// Don't forget to reset data between runs, else we're gonna stick
			// around with former data if new one isn't specified
			unset($data);
			
			foreach($item->attributes() as $key => $value)
				$data[$model->name][$key] = (string)$value;
			
			// We have CSV as some keys, expand them
			for ($i = 0; $i < $numAttribs; $i++) {
				// Skip missing attributes from a possibly previous version export
				if (! isset($data[$model->name][$attribNames[$i]]))
					continue;
				
				$list = split(',',$data[$model->name][$attribNames[$i]]);
				if ($data[$model->name][$attribNames[$i]] != "" && sizeof($list) > 0)
					$data[$aliases[$i]][$aliases[$i]] = $list;
				unset($data[$model->name][$attribNames[$i]]);
			}
			
			$model->create();
			if (!$model->save($data, false))
				return false;
				
			$imported++;
			if (!($imported%50))
				$reporter->setPartCurrent($imported);
		}
		
		$reporter->setPartCurrent($imported);
		return true;
	}
	
	/** Makes an XML attribute name from a model name.
	 *
	 * @param string $modelName The model name
	 * @return string The attribute name
	 */
	private static function HABTMModelToAttribute($modelName) {
		return Inflector::underscore(Inflector::pluralize($modelName));
	}
	
	/** Makes an XML container (for model elements) name from a model name.
	 *
	 * @param string $modelName The model name
	 * @return string The container name
	 */
	private static function modelNameToContainer($modelName) {
		return Inflector::underscore(Inflector::pluralize($modelName));
	}
	
	/** Makes an XML element name from a model name.
	 *
	 * @param string $modelName The model name
	 * @return string The container name
	 */
	private static function modelNameToElement($modelName) {
		return Inflector::underscore($modelName);
	}
}

?>