<?php
/* Copyright 2007-2009 Xavier 'Xr' Dalem
 * 
 * This file is part of Demyo. Demyo is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License version 3,
 * as published by the Free Software Foundation.
 * 
 * Demyo is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU GPLv3 for more details.
 */

class Help extends AppModel {
	public $useTable = false;
	
	private $helpIndexes = array();
	
	/**
	 * 
	 * @param $id
	 * @return array An array containing the data for that id it it exists,
	 * false if it doesn't (or if the documentation index forces fallback)
	 */
	public function getEntry($id) {
		$helpIndex = $this->getHelpIndex();
		
		// Different languages might conflict, but the local version is always
		// considered first
		if (isset($helpIndex['local-ids'][$id]))
			$id = $helpIndex['local-ids'][$id];
		
		if (isset($helpIndex['ids'][$id]['force-fallback']) && $helpIndex['ids'][$id]['force-fallback'] === true)
			return false;
		
		return $this->getEntryPlain($helpIndex, $id);
	}
	
	/**
	 * @param $id The ID (localized or not)
	 * @return array An array containing the fallback data, or false if there
	 * is none.
	 */
	public function getFallbackEntry($id) {
		$curLangHelpIndex = $this->getHelpIndex();
		
		// See if we have an alternative
		if (!isset($curLangHelpIndex['fallback']))
			return false;
		
		if (isset($curLangHelpIndex['local-ids'][$id]))
			$id = $curLangHelpIndex['local-ids'][$id];
		$fallbackHelpIndex = $this->getHelpIndex($curLangHelpIndex['fallback']);
		
		return $this->getEntryPlain($fallbackHelpIndex, $id);
	}
	
	/**
	 * @param $helpIndex The already retreived help index
	 * @param $id The non-localized ID
	 */
	private function getEntryPlain($helpIndex, $id) {
		if (!isset($helpIndex['ids'][$id]))
			return false;
		
		return array('lang' => $helpIndex['lang'], 'entry' => $helpIndex['ids'][$id]);
	}
	
	public function getHelpIndex($lang = false) {
		if ($lang === false)
			$lang = Configure::read('Demyo.lang');
			
		// Load from memory
		if (isset($this->helpIndexes[$lang]))
			return $this->helpIndexes[$lang];

		$indexFile = APP.'locale'.DS.$lang.DS.'help'.DS.'index.xml';
		
		// Load from disk cache
		$cacheFile = CACHE."help-$lang.php";
		if (file_exists($cacheFile) && filemtime($cacheFile) >= filemtime($indexFile)) {
			include($cacheFile);
			$this->helpIndexes[$lang] = $helpIndex;
			return $helpIndex;
		}

		// No cache, parse
		if (! file_exists($indexFile))
			throw new Exception(__s('Unable to open help index file at %s. Maybe the help for that language is missing.', $indexFile), 500);
		
		$indexSXML = simplexml_load_file($indexFile);
		
		$helpIndex = array(
			'lang' => $lang,
			'top-level' => array()
		);
		
		if (! empty($indexSXML['fallback']))
			$helpIndex['fallback'] = (string)$indexSXML['fallback'];
			
		$helpIndex['index-name'] = (string)$indexSXML['name'];
		
		if (isset($helpIndex['fallback']) && $lang == $helpIndex['fallback'])
			throw new Exception(__('Badly formed help index: fallback language is the same as current language.', true), 500);
		
		$previous = null;
		$next = null;
		foreach ($indexSXML->chapter as $chapter) {
			$cid = self::updateIndex($helpIndex, $chapter, null, $previous);
			array_push($helpIndex['top-level'], $cid);
			
			$helpIndex['ids'][$cid]['heirs'] = array();
			foreach ($chapter->section as $section) {
				$sid = self::updateIndex($helpIndex, $section, $cid, $previous);
				array_push($helpIndex['ids'][$cid]['heirs'], $sid);

				$helpIndex['ids'][$sid]['heirs'] = array();
				foreach ($section->subsection as $subsection) {
					$ssid = self::updateIndex($helpIndex, $subsection, $sid, $previous);
					array_push($helpIndex['ids'][$sid]['heirs'], $ssid);
				}
			}
		}
		
		
		// Save to memory and disk cache
		$this->helpIndexes[$lang] = $helpIndex;
		$string = '<?php $helpIndex = unserialize(\'';
		$string .= serialize($helpIndex);
		$string .= '\'); ?>';
		file_put_contents($cacheFile, $string);
		
		return $helpIndex;
	}
	
	private static function srcToAnchoredFile($src) {
		if ($src === '')
			return array('file' => null, 'anchor' => null);
		
		if (($pos = strpos($src, '#')) !== false)
			return array('file' => substr($src, 0, $pos), 'anchor' => substr($src, $pos+1));
		else
			return array('file' => $src, 'anchor' => null);
	}
	
	private static function updateIndex(&$index, $element, $parent, &$previous) {
		$id = (string) $element['id'];
		$index['ids'][$id]['name'] = (string) $element['name'];
		$src = str_replace('/', DS, self::srcToAnchoredFile((string) $element['src'])); // Fix directory separators
		$index['ids'][$id]['src'] = $src['file'];
		$index['ids'][$id]['anchor'] = $src['anchor'];
		if (isset($element['local-id'])) {
			$index['local-ids'][(string) $element['local-id']] = $id;
			$index['ids'][$id]['local-id'] = (string) $element['local-id'];
		}
		if (isset($element['force-fallback'])) {
			$index['ids'][$id]['force-fallback'] = (boolean)$element['force-fallback'];
		}
		$index['ids'][$id]['parent'] = $parent;
		
		// Explicit 'next'. Set it, a prepare the 'next' element for
		// the corresponding backwards jump.
		// Obviously, the last element to jump to that 'next' will be
		// its 'previous'
		if (isset($element['next'])) {
			$index['ids'][$id]['next'] = (string) $element['next'];
			$index['ids'][(string) $element['next']]['previous'] = $id;
		}
		
		// Don't override the 'previous' if it was set by a 'next'
		if (! isset($index['ids'][$id]['previous'])) {
			$index['ids'][$id]['previous'] = $previous;
		}
		
		// Don't override the 'next' if it was explicit or already set
		// by a former entry (in which case it is closer to the former
		// entry, thus preferable)
		if ($previous !== null && !isset($index['ids'][$previous]['next']) && $src['file'] !== $index['ids'][$previous]['src'])
				$index['ids'][$previous]['next'] = $id;
		
		// Always honor forced 'previous', but don't update anything in the
		// index: this is to fix cases where going up can be unintuitive
		if (isset($element['previous']))
			$index['ids'][$id]['previous'] = (string) $element['previous'];
		
		if ($src['file'] !== null)
			$previous = $id;
		if (($previous === null && $src['file'] !== null)
				|| ($previous !== null && $src['file'] !== $index['ids'][$previous]['src']))
			$previous = $id;
		
		return $id;
	}
}
?>
