<?php
/* Copyright 2007-2009 Xavier 'Xr' Dalem
 * 
 * This file is part of Demyo. Demyo is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License version 3,
 * as published by the Free Software Foundation.
 * 
 * Demyo is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU GPLv3 for more details.
 */
class Series extends AppModel {
	public $name = 'Series';
	public $order = 'name ASC';
	public $actsAs = array('Containable', 'SimpleXML' => array('useHABTM' => true));
	public $hasMany = array(
		'Album'	=> array(
			'className' => 'Album',
			'order' => 'Album.cycle ASC, Album.number ASC',
			'dependent' => true
			),
		'Derivative');
	public $hasAndBelongsToMany = array(
		'RelatedSeries' => array(
			'className' => 'Series',
			'joinTable' => 'series_relations',
			'foreignKey' => 'main',
			'associationForeignKey' => 'sub',
			'order' => 'RelatedSeries.name'
			)
		);
	
	/**
	 * Saves model data to the database. By default, validation occurs before save.
	 *
	 * @param array $data Data to save.
	 * @param boolean $validate If set, validation will be done before the save
	 * @param array $fieldList List of fields to allow to be written
	 * @return mixed On success Model::$data if its not empty or true, false on failure
	 */
	public function save($data = null, $validate = true, $fieldList = array()) {
		$main = $this->hasAndBelongsToMany['RelatedSeries']['foreignKey'];
		$sub = $this->hasAndBelongsToMany['RelatedSeries']['associationForeignKey'];
		$table = $this->tablePrefix.$this->hasAndBelongsToMany['RelatedSeries']['joinTable'];
		
		/*
		We have to edit the reversed relation, i.e. set the other series
		to have a relation with the current. The relation of the current
		series with the other ones is handled by parent::save.

		. Let C be the series currently related to this one
		. Let N be the new relations (the ones we want to save, wether they
		are actually new or were already set)
		. Then, C-N=R is the set of relations to remove and N-C=A the set of
		associations to add

		Note: We cannot call parent::save first because this would
		corrupt the computation of the C and N sets (since parent::save will
		insert a part of the series relation, which we will use later).
		However, in order to compute C, we need the id of the saved series,
		which is easy to obtain from $data if the series already existed
		but cannot be found until parent::save has been called if the $data
		to save is in fact a new series (we need lastInsertId() which will
		actually be in $this->id after parent::save has been called.
		The solution to this problem is the following
			- If $data['Series']['id'] is set, use it, compute C and call
			  parent::save
			- If it's not set, call parent::save, use $this->id and set C
			  as empty (they were previously no series attached to the new one
			  since it didn't exist)
		 */
		
		if (! empty($data['Series']['id'])) { // edition of an existing series
			if (! ctype_digit($data['Series']['id']))
				return false;
			$id = $data['Series']['id'];
			$C = Set::extract('/Relations/sub', $this->query("SELECT $sub FROM $table AS Relations WHERE $main = $id"));
			$result = parent::save($data, $validate, $fieldList);
			if (! $result)
				return $result;
		}
		
		else { // insertion of a new series
			$result = parent::save($data, $validate, $fieldList);
			if (! $result)
				return $result;
			$id = $this->id;
			$C = array();
		}
		
		// The following does not depend on the add/edit action anymore
		$N = (!empty($data['RelatedSeries']['RelatedSeries']))?$data['RelatedSeries']['RelatedSeries']:array();
		$R = array_diff($C, $N);
		$A = array_diff($N, $C);
		
		if (! empty($R))
			$this->query("DELETE FROM $table WHERE $sub = $id AND $main IN (".implode(',', $R).')');
		
		if (! empty($A)) {
			foreach ($A as &$item)
				$item = "($item, $id)";
			$this->query("INSERT INTO $table ($main, $sub) VALUES ".implode(",", $A));
		}
		
		return $result;
	}
	
	/**
	 * Removes record for given id. If no id is given, the current id is used.
	 *
	 * @param mixed $id ID of record to delete
	 * @param boolean $cascade Set to true to delete records that depend on this record
	 * @return boolean True on success
	 * @access public
	 */
	function del($id = null, $cascade = true) {
		if (! parent::del($id, $cascade))
			return false;
		
		// We also need to delete reverse relations since they were inserted
		// manually
		$sub = $this->hasAndBelongsToMany['RelatedSeries']['associationForeignKey'];
		$table = $this->tablePrefix.$this->hasAndBelongsToMany['RelatedSeries']['joinTable'];
		$this->query("DELETE FROM $table WHERE $sub = $id");
		return true;
	}
		
	public function getImages($id = null) {
		if ($id === null)
			$id = $this->id;
		
		$this->recursive = 2;
		$this->Album->unbindAll(array('belongsTo' => array('Cover'), 'hasAndBelongsToMany' => array('Image')));
		$this->Derivative->unbindAll(array('hasAndBelongsToMany' => array('Image')));
		$series = $this->read(null, $id);
		
		$images = array();
		
		// Albums
		foreach ($series['Album'] as $album)
			$images = array_merge($images, array($album['Cover']), $album['Image']);
		
		// Derivatives
		foreach ($series['Derivative'] as $deriv)
			$images = array_merge($images, $deriv['Image']);
		
		return $images;
	}
}

?>
